/*
 * Copyright 2008 Perforce Software.  All rights reserved.
 *
 * This file is part of Perforce - the FAST SCM System.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "MyDTG.h"
#include "MyDTS.h"
extern "C" {
#include <dtg-utils.h>
}
#include <dtg-str.h>

const char *MyDTGDefect::MyDTGMagic = "MyDTGDefectClass";

MyDTGDefect::MyDTGDefect( MyDTGProj *proj, 
	                const char *in_defect, 
	                struct DTGError *error  )
{
	magic = MyDTGMagic;
	in_proj = proj;
	fields = NULL;
	changes = NULL;
	defect = mk_string( in_defect );
	if( !proj || !defect )
	{
	    set_DTGError( error, 
			"MyDTGDefect::MyDTGDefect: Undefined arguments");
	    return;
	}
	dirty = !strcmp( defect, "new" );
	testing = in_proj->testing;
	if( testing )
	{
	    testing = 1;
	    clear_DTGError( error );
	    return;
	}

	char *err = NULL;
	fields = in_proj->in_dt->dts->get_defect( defect, err );
	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = in_proj->in_dt->dts->is_valid();
	    delete[] err;
	    return;
	}

	clear_DTGError( error );
}

MyDTGDefect::~MyDTGDefect()
{
	if( fields )
	    delete_DTGField( fields );
	if( changes )
	    delete_DTGField( changes );
	if( defect )
	    delete[] defect;
}

struct DTGField *MyDTGDefect::get_fields( struct DTGError *error )
{
	struct DTGField *list = NULL;
	if( testing )
	{
	    list = new_DTGField( "*name*", "*value*" );
	    clear_DTGError( error );
	    return list;
	}

	if( !fields )
	{
	    set_DTGError( error, 
			"MyDTGDefect::get_fields: Currently not implemented");
	    return NULL;
	}
	for( struct DTGField *f = fields; f; f = f->next )
	   list = append_DTGField( list, new_DTGField( f->name, f->value ) );

	clear_DTGError( error );
	return list;
}

static const char *VALUE_CONST = "*value*";

char *MyDTGDefect::get_field( const char *name, struct DTGError *error )
{
	if( testing )
	    if( name && !strcmp( name, "*name*" ) )
	    {
	        clear_DTGError( error );
	        return strdup( VALUE_CONST );
	    }
	    else
	    {
	        set_DTGError( error, 
				"MyDTGDefect::get_field: Field not defined");
	        return NULL;
	    }

	if( !fields )
	{
	    set_DTGError( error, 
				"MyDTGDefect::get_field: Fields not defined");
	    return NULL;
	}

	clear_DTGError( error );
	if( !strcmp( "Fixes", name ) )
	{
	    set_DTGError( error, "\"Fixes\" is a special field, and can only be"
	                  " the destination for \"Fix Details\"." );
	    error->can_continue = 0;
	    return NULL;
	}

	struct DTGField *f;
	for( f = changes; f; f = f->next )
	    if( !strcmp( f->name, name ) )
	        return f->value ? strdup( f->value ) : NULL;
	for( f = fields; f; f = f->next )
	    if( !strcmp( f->name, name ) )
	        return f->value ? strdup( f->value ) : NULL;
	return NULL;
}

void MyDTGDefect::set_field( const char *name, const char *value,
	                    struct DTGError *error )
{
	if( testing )
	{
	    if( name && value && !strcmp( name, "*name*" ) )
	        clear_DTGError( error );
	    else
	        set_DTGError( error, 
				"MyDTGDefect::set_field: Field not defined");
	    return;
	}

	if( !fields )
	{
	    set_DTGError( error, "MyDTGDefect::set_field: Fields not defined");
	    return;
	}

	struct DTGField *f;
	for( f = changes; f; f = f->next )
	    if( !strcmp( f->name, name ) )
	    {
	        if( f->value )
	            free( f->value );
	        f->value = strdup( value ); // struct DTGField uses free()
	        dirty = 1;
	        break;
	    }
	if( !f )
	{
	    f = new_DTGField( name, value );
	    f->next = changes;
	    changes = f;
	    dirty = 1;
	}

	clear_DTGError( error );
	return;
}

static const char *NAME_CONST = "*defect*";

static struct DTGField *find_field( const char *name, struct DTGField *fields )
{
	if( !name || !fields )
	    return NULL;
	struct DTGField *f;
	for( f=fields; f; f = f->next )
	    if( !strcmp( f->name, name ) )
	        break;
	return f;
}

char *MyDTGDefect::save( DTGError *error )
{
	if( testing )
	{
	    clear_DTGError( error );
	    return strdup( NAME_CONST );
	}

	if( !fields )
	{
	    set_DTGError( error, "MyDTGDefect::save: Fields not defined");
	    return NULL;
	}

	if( dirty )
	{
	    struct DTGField *due_date = find_field( "DueDate", changes );
	    struct DTGDate *due = NULL;
	    if( due_date && due_date->value )
	    {
	        if( *due_date->value )
	            due = in_proj->in_dt->extract_date( due_date->value );
	        else
	        {
                    free( due_date->value );
	            due_date->value = NULL;
	        }
	    }

	    struct DTGField *start_date = find_field( "StartDate", changes );
	    struct DTGDate *start = NULL;
	    if( start_date && start_date->value )
	    {
	        if( *start_date->value )
	            start = in_proj->in_dt->extract_date( start_date->value );
	        else
	        {
                    free( start_date->value );
	            start_date->value = NULL;
	        }
	    }

	    if( start && due )
	    {
	        if( compare_DTGDate( due, start ) == 1 )
	        {
	            delete_DTGDate( due );
	            delete_DTGDate( start );
	            set_DTGError( error, 
				"DueDate may not earlier than StartDate" );
	            return NULL;
	        }
	    }
	    delete_DTGDate( due );
	    delete_DTGDate( start );

	    char *err = NULL;
	    char *name = 
		in_proj->in_dt->dts->save_defect( defect, changes, err );
	    if( defect )
	        delete[] defect;
	    defect = name;
	    if( err )
	    {
	        set_DTGError( error, err );
	        error->can_continue = in_proj->in_dt->dts->is_valid();
	        delete[] err;
	    }
	    else
	        dirty = 0;
	}
	return defect ? strdup( defect ) : NULL;
}
